@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2021 salesforce.com, inc.
+# Copyright (c) 2022 salesforce.com, inc.
 # All rights reserved.
 # SPDX-License-Identifier: BSD-3-Clause
 # For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
@@ -8,11 +8,12 @@
 Base class for anomaly detectors.
 """
 from abc import abstractmethod
-from copy import copy, deepcopy
+import copy
 import inspect
 import logging
 from typing import Any, Dict, List
 
+import pandas as pd
 from scipy.stats import norm
 
 from merlion.models.base import Config, ModelBase, MultipleTimeseriesModelMixin
@@ -21,7 +22,7 @@ from merlion.post_process.calibrate import AnomScoreCalibrator
 from merlion.post_process.factory import PostRuleFactory
 from merlion.post_process.sequence import PostRuleSequence
 from merlion.post_process.threshold import AggregateAlarms, Threshold
-from merlion.utils import TimeSeries
+from merlion.utils import TimeSeries, UnivariateTimeSeries
 
 logger = logging.getLogger(__name__)
 
@@ -34,6 +35,8 @@ class DetectorConfig(Config):
     _default_threshold = AggregateAlarms(alm_threshold=3.0)
     calibrator: AnomScoreCalibrator = None
     threshold: Threshold = None
+    enable_calibrator: bool = True
+    enable_threshold: bool = True
 
     def __init__(
         self, max_score: float = 1000, threshold=None, enable_calibrator=True, enable_threshold=True, **kwargs
@@ -54,7 +57,7 @@ class DetectorConfig(Config):
         self.enable_calibrator = enable_calibrator
         self.calibrator = AnomScoreCalibrator(max_score=max_score)
         if threshold is None:
-            self.threshold = deepcopy(self._default_threshold)
+            self.threshold = copy.deepcopy(self._default_threshold)
         elif isinstance(threshold, dict):
             self.threshold = PostRuleFactory.create(**threshold)
         else:
@@ -174,7 +177,6 @@ class DetectorBase(ModelBase):
     def post_rule(self):
         return self.config.post_rule
 
-    @abstractmethod
     def train(
         self, train_data: TimeSeries, anomaly_labels: TimeSeries = None, train_config=None, post_rule_train_config=None
     ) -> TimeSeries:
@@ -194,13 +196,23 @@ class DetectorBase(ModelBase):
         :return: A `TimeSeries` of the model's anomaly scores on the training
             data.
         """
+        if train_config is None:
+            train_config = copy.deepcopy(self._default_train_config)
+        train_data = self.train_pre_process(train_data).to_pd()
+        train_scores = self._train(train_data=train_data, train_config=train_config)
+        train_scores = UnivariateTimeSeries.from_pd(train_scores, name="anom_score").to_ts()
+        self.train_post_rule(train_scores, anomaly_labels, post_rule_train_config)
+        return train_scores
+
+    @abstractmethod
+    def _train(self, train_data: pd.DataFrame, train_config=None) -> pd.DataFrame:
         raise NotImplementedError
 
     def train_post_rule(
         self, anomaly_scores: TimeSeries, anomaly_labels: TimeSeries = None, post_rule_train_config=None
     ):
         if self.post_rule is not None:
-            kwargs = copy(self._default_post_rule_train_config)
+            kwargs = copy.copy(self._default_post_rule_train_config)
             if post_rule_train_config is not None:
                 kwargs.update(post_rule_train_config)
             params = inspect.signature(self.post_rule.train).parameters
@@ -209,6 +221,9 @@ class DetectorBase(ModelBase):
             self.post_rule.train(anomaly_scores=anomaly_scores, anomaly_labels=anomaly_labels, **kwargs)
 
     @abstractmethod
+    def _get_anomaly_score(self, time_series: pd.DataFrame, time_series_prev: pd.DataFrame = None) -> pd.DataFrame:
+        raise NotImplementedError
+
     def get_anomaly_score(self, time_series: TimeSeries, time_series_prev: TimeSeries = None) -> TimeSeries:
         """
         Returns the model's predicted sequence of anomaly scores.
@@ -221,7 +236,31 @@ class DetectorBase(ModelBase):
             immediately follows the training data.
         :return: a univariate `TimeSeries` of anomaly scores
         """
-        raise NotImplementedError
+        # Ensure the dimensions are correct
+        assert (
+            time_series.dim == self.dim
+        ), f"Expected time_series to have dimension {self.dim}, but got {time_series.dim}."
+        if time_series_prev is not None:
+            assert (
+                time_series_prev.dim == self.dim
+            ), f"Expected time_series_prev to have dimension {self.dim}, but got {time_series_prev.dim}."
+
+        # Transform the time series
+        time_series, time_series_prev = self.transform_time_series(time_series, time_series_prev)
+        if self.require_univariate:
+            assert time_series.dim == 1, (
+                f"{type(self).__name__} model only accepts univariate time series, but time series "
+                f"(after transform {self.transform}) has dimension {time_series.dim}."
+            )
+
+        time_series = time_series.to_pd()
+        if time_series_prev is not None:
+            time_series_prev = time_series_prev.to_pd()
+
+        # Get the anomaly scores & ensure the dimensions are correct
+        anom_scores = self._get_anomaly_score(time_series, time_series_prev)
+        assert anom_scores.shape[1] == 1, f"Expected anomaly scores returned by {type(self)} to be univariate."
+        return UnivariateTimeSeries.from_pd(anom_scores, name="anom_score").to_ts()
 
     def get_anomaly_label(self, time_series: TimeSeries, time_series_prev: TimeSeries = None) -> TimeSeries:
         """
@@ -358,10 +397,14 @@ class MultipleTimeseriesDetectorMixin(MultipleTimeseriesModelMixin):
     """
     Abstract mixin for anomaly detectors supporting training on multiple time series.
     """
+
     @abstractmethod
     def train_multiple(
-        self, multiple_train_data: List[TimeSeries], anomaly_labels: List[TimeSeries] = None,
-        train_config=None, post_rule_train_config=None
+        self,
+        multiple_train_data: List[TimeSeries],
+        anomaly_labels: List[TimeSeries] = None,
+        train_config=None,
+        post_rule_train_config=None,
     ) -> List[TimeSeries]:
         """
         Trains the anomaly detector (unsupervised) and its post-rule
