@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2021 salesforce.com, inc.
+# Copyright (c) 2022 salesforce.com, inc.
 # All rights reserved.
 # SPDX-License-Identifier: BSD-3-Clause
 # For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
@@ -21,7 +21,7 @@ from scipy.special import logsumexp
 from scipy.stats import norm
 from tqdm import tqdm
 
-from merlion.models.anomaly.base import NoCalibrationDetectorConfig
+from merlion.models.anomaly.base import NoCalibrationDetectorConfig, DetectorBase
 from merlion.models.anomaly.forecast_based.base import ForecastingDetectorBase
 from merlion.models.forecast.base import ForecasterConfig
 from merlion.plot import Figure
@@ -158,6 +158,18 @@ class BOCPD(ForecastingDetectorBase):
         self.full_run_length_posterior = scipy.sparse.dok_matrix((0, 0), dtype=float)
         self.pw_model: List[Tuple[pd.Timestamp, ConjPrior]] = []
 
+    @property
+    def _online_model(self) -> bool:
+        return True
+
+    @property
+    def require_even_sampling(self) -> bool:
+        return False
+
+    @property
+    def require_univariate(self) -> bool:
+        return False
+
     @property
     def last_train_time(self):
         return None if len(self.train_timestamps) == 0 else to_pd_datetime(self.train_timestamps[-1])
@@ -258,9 +270,7 @@ class BOCPD(ForecastingDetectorBase):
             _, data = train_data.bisect(t0, t_in_left=False)
             self.pw_model.append((t0, self.change_kind.value(data)))
 
-    def train_pre_process(
-        self, train_data: TimeSeries, require_even_sampling: bool, require_univariate: bool
-    ) -> TimeSeries:
+    def train_pre_process(self, train_data: TimeSeries) -> TimeSeries:
         # BOCPD doesn't _require_ target_seq_index to be specified, but train_pre_process() does.
         if self.target_seq_index is None and train_data.dim > 1:
             self.config.target_seq_index = 0
@@ -268,26 +278,17 @@ class BOCPD(ForecastingDetectorBase):
                 f"Received a {train_data.dim}-variate time series, but `target_seq_index` was not "
                 f"specified. Setting `target_seq_index = 0` so the `forecast()` method will work."
             )
-        train_data = super().train_pre_process(train_data, require_even_sampling, require_univariate)
+        train_data = super().train_pre_process(train_data)
         # We manually update self.train_data in update(), so do nothing here
         self.train_data = None
         return train_data
 
-    def forecast(
-        self,
-        time_stamps: Union[int, List[int]],
-        time_series_prev: TimeSeries = None,
-        return_iqr: bool = False,
-        return_prev: bool = False,
-    ) -> Union[Tuple[TimeSeries, TimeSeries], Tuple[TimeSeries, TimeSeries, TimeSeries]]:
-        if time_series_prev is not None:
-            self.update(time_series_prev)
-        if isinstance(time_stamps, (int, float)):
-            time_stamps = pd.date_range(start=self.last_train_time, freq=self.timedelta, periods=int(time_stamps))[1:]
-        else:
-            time_stamps = to_pd_datetime(time_stamps)
+    def _forecast(
+        self, time_stamps: List[int], time_series_prev: pd.DataFrame = None, return_prev=False
+    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
+        time_stamps = to_pd_datetime(time_stamps)
         if return_prev and time_series_prev is not None:
-            time_stamps = to_pd_datetime(time_series_prev.time_stamps).union(time_stamps)
+            time_stamps = time_series_prev.index.union(time_stamps)
 
         # Initialize output accumulators
         pred_full, err_full = None, None
@@ -320,37 +321,7 @@ class BOCPD(ForecastingDetectorBase):
         pred[pred.isna() | np.isinf(pred)] = 0
         err[err.isna() | np.isinf(err)] = 0
 
-        if return_iqr:
-            name = pred.name
-            lb = UnivariateTimeSeries.from_pd(pred + norm.ppf(0.25) * err, name=f"{name}_lower")
-            ub = UnivariateTimeSeries.from_pd(pred + norm.ppf(0.75) * err, name=f"{name}_upper")
-            return TimeSeries.from_pd(pred), lb.to_ts(), ub.to_ts()
-        return TimeSeries.from_pd(pred), TimeSeries.from_pd(err)
-
-    def get_figure(
-        self,
-        *,
-        time_series: TimeSeries = None,
-        time_stamps: List[int] = None,
-        time_series_prev: TimeSeries = None,
-        plot_anomaly=True,
-        filter_scores=True,
-        plot_forecast=False,
-        plot_forecast_uncertainty=False,
-        plot_time_series_prev=False,
-    ) -> Figure:
-        if time_series is not None:
-            self.update(time_series)
-        return super().get_figure(
-            time_series=time_series,
-            time_stamps=time_stamps,
-            time_series_prev=time_series_prev,
-            plot_anomaly=plot_anomaly,
-            filter_scores=filter_scores,
-            plot_forecast=plot_forecast,
-            plot_forecast_uncertainty=plot_forecast_uncertainty,
-            plot_time_series_prev=plot_time_series_prev,
-        )
+        return pd.DataFrame(pred), pd.DataFrame(err)
 
     def update(self, time_series: TimeSeries):
         """
@@ -362,9 +333,7 @@ class BOCPD(ForecastingDetectorBase):
         # Only update on the portion of the time series after the last training timestamp
         time_stamps = time_series.time_stamps
         if self.last_train_time is not None:
-            time_series_prev, time_series = time_series.bisect(self.last_train_time, t_in_left=True)
-        else:
-            time_series_prev = None
+            _, time_series = time_series.bisect(self.last_train_time, t_in_left=True)
 
         # Update the training data accumulated so far
         if self.train_data is None:
@@ -372,9 +341,6 @@ class BOCPD(ForecastingDetectorBase):
         else:
             self.train_data = self.train_data + time_series
 
-        # Apply any pre-processing transforms to the time series
-        time_series, time_series_prev = self.transform_time_series(time_series, time_series_prev)
-
         # Align the time series & expand the array storing the full posterior distribution of run lengths
         time_series = time_series.align()
         n_seen, T = self.n_seen, len(time_series)
@@ -448,40 +414,57 @@ class BOCPD(ForecastingDetectorBase):
         # Return the anomaly scores
         return self._get_anom_scores(time_stamps)
 
-    def train(
-        self, train_data: TimeSeries, anomaly_labels: TimeSeries = None, train_config=None, post_rule_train_config=None
-    ) -> TimeSeries:
+    def _train(self, train_data: pd.DataFrame, train_config=None) -> pd.DataFrame:
+        # If not automatically detecting the change kind, train as normal
+        if self.change_kind is not ChangeKind.Auto:
+            return self.update(time_series=TimeSeries.from_pd(train_data)).to_pd()
 
-        # If automatically detecting the change kind, train candidate models with each change kind
-        # TODO: abstract this logic into a merlion.models.automl.GridSearch object?
-        if self.change_kind is ChangeKind.Auto:
-            candidates = []
-            for change_kind in ChangeKind:
-                if change_kind is ChangeKind.Auto:
-                    continue
-                candidate = copy.deepcopy(self)
-                candidate.config.change_kind = change_kind
-                train_scores = candidate.train(train_data, anomaly_labels, train_config, post_rule_train_config)
-                log_likelihood = logsumexp([p.logp for p in candidate.posterior_beam])
-                candidates.append((candidate, train_scores, log_likelihood))
-                logger.info(f"Change kind {change_kind.name} has log likelihood {log_likelihood:.3f}.")
+        # Otherwise, evaluate all change kinds as options
+        candidates = []
+        for change_kind in [ck for ck in ChangeKind if ck is not ChangeKind.Auto]:
+            candidate = copy.deepcopy(self)
+            candidate.config.change_kind = change_kind
+            train_scores = candidate._train(train_data, train_config=train_config)
+            log_likelihood = logsumexp([p.logp for p in candidate.posterior_beam]).item()
+            candidates.append((log_likelihood, candidate, train_scores))
+            logger.info(f"Change kind {change_kind.name} has log likelihood {log_likelihood:.3f}.")
 
-            # Choose the model with the best log likelihood
-            i_best = np.argmax([candidate[2] for candidate in candidates])
-            best, train_scores, _ = candidates[i_best]
-            self.__setstate__(best.__getstate__())
-            logger.info(f"Using change kind {self.change_kind.name} because it has the best log likelihood.")
-
-        # Otherwise, just train as normal
-        else:
-            self.train_pre_process(train_data, require_even_sampling=False, require_univariate=False)
-            train_scores = self.update(time_series=train_data)
-            self.train_post_rule(train_scores, anomaly_labels, post_rule_train_config)
-
-        # Return the anomaly scores on the training data
+        # Choose the model with the best log likelihood
+        i_best = np.argmax([candidate[0] for candidate in candidates])
+        log_likelihood, best, train_scores = candidates[i_best]
+        self.__setstate__(best.__getstate__())
+        logger.info(f"Using change kind {self.change_kind.name} because it has the best log likelihood.")
         return train_scores
 
     def get_anomaly_score(self, time_series: TimeSeries, time_series_prev: TimeSeries = None) -> TimeSeries:
+        return DetectorBase.get_anomaly_score(self, time_series, time_series_prev)
+
+    def _get_anomaly_score(self, time_series: pd.DataFrame, time_series_prev: pd.DataFrame = None) -> pd.DataFrame:
         if time_series_prev is not None:
-            self.update(time_series_prev)
-        return self.update(time_series)
+            self.update(TimeSeries.from_pd(time_series_prev))
+        return self.update(TimeSeries.from_pd(time_series)).to_pd()
+
+    def get_figure(
+        self,
+        *,
+        time_series: TimeSeries = None,
+        time_stamps: List[int] = None,
+        time_series_prev: TimeSeries = None,
+        plot_anomaly=True,
+        filter_scores=True,
+        plot_forecast=False,
+        plot_forecast_uncertainty=False,
+        plot_time_series_prev=False,
+    ) -> Figure:
+        if time_series is not None:
+            self.update(self.transform(time_series))
+        return super().get_figure(
+            time_series=time_series,
+            time_stamps=time_stamps,
+            time_series_prev=time_series_prev,
+            plot_anomaly=plot_anomaly,
+            filter_scores=filter_scores,
+            plot_forecast=plot_forecast,
+            plot_forecast_uncertainty=plot_forecast_uncertainty,
+            plot_time_series_prev=plot_time_series_prev,
+        )
